#!/usr/bin/env python

"""
Produce syspython code from an inspected program.

Copyright (C) 2006, 2007, 2010, 2011, 2012, 2013 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from micropython.common import *
from micropython.data import *
from micropython.errors import *
from os.path import exists, extsep, join
import compiler.ast
import sys
import os

try:
    set
except NameError:
    from sets import Set as set

# Convenience definitions.

module_attribute = compiler.ast.Getattr
special_name = compiler.ast.Name

def quoted_ref(obj):
    return compiler.ast.CallFunc("__static__", [compiler.ast.Const(obj.full_name())])

def quoted_name(s):
    return compiler.ast.Const(s)

# Special function names.
# NOTE: Some of the assignment operations should not be supported unless
# NOTE: attribute usage observations are being made.

assattr_functions = ("__storeattrcontext__", "__storeattrcontext__",     "__storeattr__",
                     "__storeattrindex__",   None)
getattr_functions = ("__loadattrcontext__",  "__loadattrcontextcond__",  "__loadattr__",
                     "__loadattrindex__",    "__loadattrindexcontextcond__")

# Source code classes.

class ConvertedSource(ASTVisitor):

    "A conversion of module source code to syspython."

    def __init__(self, module, program):
        self.visitor = self
        self.module = module
        self.program = program
        self.objtable = program.get_object_table()
        self.in_main = False
        self.units = []

    def get_unit(self):
        return self.units[-1]

    def get_module(self):
        return self.units[0]

    def to_stream(self, stream):

        "Write the converted code to the given 'stream'."

        module = self.dispatch(self.module.astnode)
        stream.write(str(module))

    def NOP(self, node):
        return node

    def visitModule(self, node):
        module = node.unit
        self.units.append(module)

        definitions = self.process_definitions(node)

        # __globalnames__(name, ...)

        globalnames = module.module_attribute_names() and [
            compiler.ast.CallFunc(
                special_name("__globalnames__"),
                [special_name(name) for name in module.module_attribute_names()]
                )
            ] or []

        # def __main__():
        #     ...

        self.in_main = True

        main = compiler.ast.Function(
            [], "__main__", [], [], 0, "Module initialisation.",
            compiler.ast.Stmt(globalnames + self.dispatch(node.node).nodes)
            )

        self.in_main = False
        self.units.pop()

        return compiler.ast.Module(node.doc, compiler.ast.Stmt(definitions + [main]))

    # Statements.

    def visitAssert(self, node):
        return compiler.ast.Assert(self.dispatch(node.test), node.fail and self.dispatch(node.fail))

    def visitAssign(self, node):
        expr = self.dispatch(node.expr)
        return compiler.ast.Stmt([self.dispatch(n, expr) for n in node.nodes])

    def visitAugAssign(self, node):

        #    lvalue = op(lvalue, expr)
        # -> __fn__(lvalue, op(lvalue, expr))

        op_name = operator_functions[node.op]

        return self.dispatch(node.node, compiler.ast.CallFunc(
            module_attribute("operator", op_name),
            [self.dispatch(node.node), self.dispatch(node.expr)]
            ))

    visitBreak = NOP

    def visitClass(self, node):
        if not used_by_unit(node):
            return compiler.ast.Stmt([])

        self.units.append(node.unit)
        try:
            # Incorporate class body code in the main function.

            if self.in_main:
                return self.dispatch(node.code)
            else:
                return self._visitClassDefinition(node)

        finally:
            self.units.pop()

    def _visitClassDefinition(self, node):
        cls = node.unit

        # __instattrs__(name, ...)
        # __clsattrs__(name, ...)

        instattrs = cls.instance_attribute_names() and [
            compiler.ast.CallFunc(
                special_name("__instattrs__"),
                [special_name(name) for name in cls.instance_attribute_names()]
                )
            ] or []

        clsattrs = cls.class_attribute_names() and [
            compiler.ast.CallFunc(
                special_name("__clsattrs__"),
                [special_name(name) for name in cls.class_attribute_names()]
                )
            ] or []

        # __inherited__(superclass, name, ...)
        # ...

        attrs_by_cls = {}
        for attrname, attr in cls.all_class_attributes().items():
            supercls = attr.parent
            if supercls is cls:
                continue
            if not attrs_by_cls.has_key(supercls):
                attrs_by_cls[supercls] = []
            attrs_by_cls[supercls].append(attrname)

        inherited = []

        for supercls, attrnames in attrs_by_cls.items():
            inherited.append(
                compiler.ast.CallFunc(
                    special_name("__inherited__"),
                    [quoted_ref(supercls)] + [special_name(name) for name in attrnames]
                    ))

        # __descendants__(name, ...)

        descendants = cls.all_descendants() and [
            compiler.ast.CallFunc(
                special_name("__descendants__"),
                [special_name(name) for name in cls.all_descendants().keys()]
                )
            ] or []

        # Process all the definitions defined inside the class.

        definitions = self.process_definitions(node)

        return compiler.ast.Class(node.name, [], node.doc,
            compiler.ast.Stmt(instattrs + clsattrs + inherited + descendants + definitions)
            )

    visitContinue = NOP

    def visitDiscard(self, node):
        return compiler.ast.Discard(self.dispatch(node.expr))

    def visitFor(self, node):

        """
        Convert from...

        for <assign> in <list>:
            <body>
        [ else:
            <else_> ]

        ...to...

        _it = iter(<list>)
        while True:
            try:
                <assign> = _it.next()
            except StopIteration:
                [ <else_> ]
                break
            else:
                <body>
        """

        unit = self.get_unit()
        temp = quoted_name(unit.temp_usage)
        unit.temp_usage += 1

        else_nodes = node.else_ and self.dispatch(node.else_).nodes or []

        return compiler.ast.Stmt([
            # __storetemp__(_it, __loadattr__(__builtins__, iter)(<list>))
            compiler.ast.CallFunc(special_name("__storetemp__"), [
                temp,
                compiler.ast.CallFunc(
                    compiler.ast.CallFunc(special_name("__loadattr__"),
                        [special_name("__builtins__"), special_name("iter")]
                        ),
                    [self.dispatch(node.list)]
                    )
                ]),
            # while True: ...
            compiler.ast.While(
                special_name("True"),
                # try: ...
                compiler.ast.TryExcept(
                    compiler.ast.Stmt([
                        # <assign> = ...
                        self.dispatch(node.assign,
                            # _it.next()
                            compiler.ast.CallFunc(
                                compiler.ast.CallFunc(special_name("__loadattr__"), [
                                    compiler.ast.CallFunc(special_name("__loadtemp__"), [temp]),
                                    special_name("next")
                                    ]),
                                []
                                )
                            )
                        ]),
                    # except StopIteration: ...
                    [(special_name("StopIteration"), None, compiler.ast.Stmt(else_nodes + [compiler.ast.Break()]))],
                    # else: ...
                    self.dispatch(node.body)
                    ),
                None
                )
            ])

    def visitFrom(self, node):

        # Generate __main__ function calls for each step in the imported module
        # hierarchy.

        statements = []

        for modname in self.module.get_module_paths(node.modname):
            statements.append(
                compiler.ast.CallFunc(special_name("%s.__main__" % modname ), [])
                )

        for name, alias in node.names:
            statements.append(
                compiler.ast.Assign(
                    [special_name(alias or name)],
                    compiler.ast.CallFunc(
                        special_name("__loadattribute__"),
                        [special_name(node.modname), special_name(name)]
                        )
                    )
                )

        return compiler.ast.Stmt(statements)

    def visitFunction(self, node):
        if not used_by_unit(node):
            return compiler.ast.Stmt([])

        self.units.append(node.unit)

        try:
            # Ignore functions when generating the main function.

            if self.in_main:
                return compiler.ast.Stmt([])
            else:
                return self._visitFunctionDefinition(node)
        finally:
            self.units.pop()

    def _visitFunctionDefinition(self, node):
        fn = node.unit

        # __localnames__(name, ...)
        # __globalnames__(name, ...)

        localnames = fn.all_locals() and [
            compiler.ast.CallFunc(
                special_name("__localnames__"),
                [special_name(name) for name in fn.all_locals().keys()]
                )
            ] or []

        globalnames = fn.globals and [
            compiler.ast.CallFunc(
                special_name("__globalnames__"),
                [special_name(name) for name in fn.globals]
                )
            ] or []

        defaults = [self.dispatch(n) for n in node.defaults]

        # NOTE: Should generate guards for attribute usage operations.

        code = self.dispatch(node.code)

        return compiler.ast.Function(node.decorators, node.name, node.argnames, defaults, node.flags, node.doc,
            compiler.ast.Stmt(localnames + globalnames + code.nodes))

    visitGlobal = NOP

    def visitIf(self, node):
        return compiler.ast.If(
            [(self.dispatch(compare), self.dispatch(stmt)) for (compare, stmt) in node.tests],
            node.else_ and self.dispatch(node.else_)
            )

    def visitImport(self, node):

        # Generate __main__ function calls for each step in the imported module
        # hierarchy.

        statements = []

        for name, alias in node.names:
            for modname in self.module.get_module_paths(name):
                statements.append(
                    compiler.ast.CallFunc(compiler.ast.Getattr(modname, "__main__"), [])
                    )

            statements.append(
                compiler.ast.Assign(
                    [special_name(alias or name.split(".")[0])],
                    compiler.ast.CallFunc(
                        special_name("__static__"),
                        [special_name(name)]
                        )
                    )
                )

        return compiler.ast.Stmt(statements)

    def visitPass(self, node):
        if not isinstance(self.get_unit(), Class):
            return compiler.ast.Pass()
        else:
            return compiler.ast.Stmt([])

    def visitPrint(self, node):
        return compiler.ast.Print(
            [self.dispatch(n) for n in node.nodes],
            node.dest and self.dispatch(node.dest)
            )

    def visitPrintnl(self, node):
        return compiler.ast.Print(
            [self.dispatch(n) for n in node.nodes],
            node.dest and self.dispatch(node.dest)
            )

    def visitRaise(self, node):
        return compiler.ast.Raise(
            node.expr1 and self.dispatch(node.expr1),
            node.expr2 and self.dispatch(node.expr2),
            node.expr3 and self.dispatch(node.expr3)
            )

    def visitReturn(self, node):
        return compiler.ast.Return(self.dispatch(node.value))

    def visitStmt(self, node):
        return compiler.ast.Stmt([self.dispatch(n) for n in node.nodes])

    def visitTryExcept(self, node):
        # NOTE: Need to dispatch to the assignment with the exception.
        return compiler.ast.TryExcept(
            self.dispatch(node.body),
            [(spec and self.dispatch(spec), assign and self.dispatch(assign), self.dispatch(statement))
                for spec, assign, statement in node.handlers],
            node.else_ and self.dispatch(node.else_)
            )

    def visitTryFinally(self, node):
        return compiler.ast.TryFinally(
            self.dispatch(node.body),
            self.dispatch(node.final)
            )

    def visitWhile(self, node):
        return compiler.ast.While(
            self.dispatch(node.test),
            self.dispatch(node.body),
            node.else_ and self.dispatch(node.else_)
            )

    def visitYield(self, node):
        return compiler.ast.Yield(self.dispatch(node.value))

    # Expression-related helper methods.

    def _visitBitBinary(self, node):
        op_name = operator_functions[node.__class__.__name__]
        last = self.dispatch(node.nodes[0])

        for n in node.nodes[1:]:
            last = compiler.ast.CallFunc(
                module_attribute("operator", op_name),
                [last, self.dispatch(n)]
                )

        return last

    def _visitBinary(self, node):
        op_name = operator_functions[node.__class__.__name__]

        return compiler.ast.CallFunc(
            module_attribute("operator", op_name),
            [self.dispatch(node.left), self.dispatch(node.right)]
            )

    def _visitUnary(self, node):
        op_name = operator_functions[node.__class__.__name__]

        return compiler.ast.CallFunc(
            module_attribute("operator", op_name),
            [self.dispatch(node.expr)]
            )

    def _generateValue(self, value):

        # Literal constants.

        if isinstance(value, Const):
            return compiler.ast.Const(value.get_value())

        # Other constant structures.

        if isinstance(value, Constant):
            return quoted_ref(value)

        return None

    def _visitAttr(self, node, expr=None):
        unit = self.get_unit()

        # Choose the appropriate special functions.

        (opattrcontext, opattrcontextcond, opattr,
            opattrindex, opattrindexcontextcond) = expr and assattr_functions or getattr_functions

        accessor = self.dispatch(node.expr)

        # Generate already-deduced accesses.

        if node._access_type == "constant":
            return self._generateValue(node._value_deduced)

        # Generate accesses via static objects and instances.

        if node._attr_deduced:
            if node._set_context == "set":
                op = opattrcontext
            elif node._set_context == "cond":
                op = opattrcontextcond
            else:
                op = opattr

            # Handle unsupported operations.

            if not op:
                raise TranslateError("Storing of class attribute %r via self not permitted." % node.attrname)

            # Define the arguments: accessor, attribute name and optional value.

            args = [
                node._access_type == "static" and \
                    self._generateValue(node._attr_deduced.parent) or accessor,
                special_name(node.attrname)
                ]

            if expr:
                args.append(expr)

            return compiler.ast.CallFunc(special_name(op), args)

        # Positioned accesses are normal accesses via instances.

        if node._access_type == "positioned":
            args = [accessor, special_name(node.attrname)]
            if expr:
                args.append(expr)
            return compiler.ast.CallFunc(special_name(opattr), args)

        # With no usable deductions, generate a table-based access.

        args = [accessor, special_name(node.attrname)]
        if expr:
            args.append(expr)
        access = compiler.ast.CallFunc(special_name(opattrindexcontextcond), args)

        # class.__class__ => __builtins__.type

        if node.attrname == "__class__":

            # __storetemp__(n, <accessor>)
            # __isclass__(n) and __loadattr__(__builtins__, type) or <access>

            temp = quoted_name(unit.temp_usage)
            unit.temp_usage += 1

            return compiler.ast.Stmt([
                compiler.ast.CallFunc(special_name("__storetemp__"), [temp, access]),
                compiler.ast.Or([
                    compiler.ast.And([
                        compiler.ast.CallFunc(
                            special_name("__isclass__"),
                            [compiler.ast.CallFunc(special_name("__loadtemp__"), [temp])]
                            ),
                        compiler.ast.CallFunc(
                            special_name("__loadattr__"),
                            [special_name("__builtins__"), special_name("type")]
                            )
                        ]),
                        access
                    ])
                ])

        # General accesses.

        else:
            return access

    # Expressions.

    def visitAdd(self, node):
        return self._visitBinary(node)

    def visitAnd(self, node):
        return compiler.ast.And([self.dispatch(n) for n in node.nodes])

    def visitAssAttr(self, node, expr=None):

        # Handle deletion.

        if compiler.ast.is_deletion(node):
            return compiler.ast.Stmt([])

        return self._visitAttr(node, expr)

    def visitAssList(self, node, expr=None):

        # Handle deletion.

        if compiler.ast.is_deletion(compiler.ast.flatten_assignment(node)):
            return compiler.ast.Stmt([])

        return compiler.ast.Stmt([
            self.dispatch(n, compiler.ast.CallFunc(
                module_attribute("operator", "getitem"),
                [expr, i]
                ))
            for (i, n) in enumerate(node.nodes)
            ])

    def visitAssName(self, node, expr=None):

        # Handle deletion.

        if compiler.ast.is_deletion(node):
            return compiler.ast.Stmt([])

        unit = self.get_unit()

        # Generate appropriate name access operation.
        # NOTE: Should generate guards for attribute usage operations.

        scope = getattr(node, "_scope", None)
        if not scope:
            attr, scope, from_name = self.get_unit()._get_with_scope(node.name)

        if scope == "constant":
            return node
        elif scope == "local":

            # Function locals are stored using a function.

            if isinstance(unit, Function):
                return compiler.ast.CallFunc(
                    special_name("__storelocal__"),
                    [special_name(node.name), expr]
                    )

            # Class locals are class attribute references.

            elif isinstance(unit, Class):
                return compiler.ast.CallFunc(
                    special_name("__storeattrcontext__"),
                    [quoted_ref(unit), special_name(node.name), expr]
                    )

            # Module locals are module attribute references.

            elif isinstance(unit, Module):
                return compiler.ast.CallFunc(
                    special_name("__storeattr__"),
                    [quoted_ref(unit), special_name(node.name), expr]
                    )
            else:
                raise TranslateError("Program unit has no local %r." % name)

        elif scope == "global":

            # Globals are references to module attributes.

            return compiler.ast.CallFunc(
                special_name("__storeattr__"),
                [quoted_ref(self.get_module()), special_name(node.name), expr]
                )

        elif scope == "builtin":

            # Builtins are accessed via the __builtins__ module.

            return compiler.ast.CallFunc(
                special_name("__storeattr__"),
                [special_name("__builtins__"), special_name(node.name), expr]
                )

        else:
            # NOTE: This may happen because a class attribute is optimised away.
            return compiler.ast.CallFunc(
                special_name("__storeunknown__"),
                [special_name(node.name), expr]
                )

    visitAssTuple = visitAssList

    def visitBitand(self, node):
        self._visitBitBinary(node)

    def visitBitor(self, node):
        self._visitBitBinary(node)

    def visitBitxor(self, node):
        self._visitBitBinary(node)

    def visitCallFunc(self, node):
        return compiler.ast.CallFunc(
            self.dispatch(node.node),
            [self.dispatch(arg) for arg in node.args],
            node.star_args and self.dispatch(node.star_args),
            node.dstar_args and self.dispatch(node.dstar_args)
            )

    def visitCompare(self, node):
        nodes = []
        left = node.expr
        for op_name, right in node.ops:
            nodes.append(
                compiler.ast.CallFunc(
                    module_attribute("operator", operator_functions.get(op_name)),
                    [self.dispatch(left), self.dispatch(right)]
                    )
                )
            left = right
        return compiler.ast.And(nodes)

    visitConst = NOP

    def visitDict(self, node):
        return compiler.ast.Dict([(self.dispatch(key), self.dispatch(value)) for (key, value) in node.items])

    def visitDiv(self, node):
        return self._visitBinary(node)

    def visitFloorDiv(self, node):
        return self._visitBinary(node)

    def visitGetattr(self, node, expr=None):
        return self._visitAttr(node, expr)

    def visitGenExpr(self, node):
        return compiler.ast.GenExpr(self.dispatch(node.code))

    def visitGenExprFor(self, node):
        expr = self.dispatch(node.iter)
        return compiler.ast.GenExprFor(
            self.dispatch(node.assign, expr), # NOTE: Needs to dispatch to AssName/AssTuple/AssList with an expression.
            expr,
            [self.dispatch(n) for n in node.ifs]
            )

    def visitGenExprIf(self, node):
        return compiler.ast.GenExprIf(self.dispatch(node.test))

    def visitGenExprInner(self, node):
        return compiler.ast.GenExprInner(
            self.dispatch(node.expr),
            [self.dispatch(n) for n in node.quals]
            )

    def visitIfExp(self, node):
        return compiler.ast.IfExp(
            self.dispatch(node.then),
            self.dispatch(node.test),
            self.dispatch(node.else_)
            )

    def visitInvert(self, node):
        return self._visitUnary(node)

    def visitKeyword(self, node):
        return compiler.ast.Keyword(
            node.name,
            self.dispatch(node.expr)
            )

    def visitLambda(self, node):
        self.units.append(node.unit)

        try:
            return compiler.ast.Lambda(
                node.argnames,
                [self.dispatch(n) for n in node.defaults],
                node.flags,
                self.dispatch(node.code)
                )
        finally:
            self.units.pop()

    def visitLeftShift(self, node):
        return self._visitBinary(node)

    def visitList(self, node, expr=None):
        if expr:
            return self.visitAssList(node, expr)
        return compiler.ast.List([self.dispatch(n) for n in node.nodes])

    def visitListComp(self, node):

        """
        Convert from...

        [<expr> for <assign> in <list> [ for <assign> in <list> ]... [ if <test> ]... ]

        ...to...

        _out = []
        ...
        """

        unit = self.get_unit()
        temp = quoted_name(unit.temp_usage)
        unit.temp_usage += 1

        return compiler.ast.Stmt([
            # __storetemp__(_out, __loadattr__(__builtins__, list)())
            compiler.ast.CallFunc(special_name("__storetemp__"), [
                temp,
                compiler.ast.CallFunc(
                    compiler.ast.CallFunc(special_name("__loadattr__"),
                        [special_name("__builtins__"), special_name("list")]
                        ),
                    []
                    )
                ]),
            # ...
            self.dispatch(node.quals[0], temp, node.expr, node.quals[1:])
            ])

    def visitListCompFor(self, node, out_temp, expr, quals):

        """
        Convert from...

        [<expr> for <assign> in <list> ...]

        ...to...

        _it = iter(<list>)
        while True:
            try:
                <assign> = _it.next()
            except StopIteration:
                break
            else:
                ...
                _out.append(<expr>)
        """

        unit = self.get_unit()
        temp = quoted_name(unit.temp_usage)
        unit.temp_usage += 1

        # Either generate more "for" or "if" constructs.

        if node.ifs or quals:
            nodes = node.ifs + quals
            body = self.dispatch(nodes[0], out_temp, expr, nodes[1:])

        # Or generate the append statement.

        else:
            body = self._visitListCompExpr(out_temp, expr)

        # Wrap the above body in the loop construct.

        return compiler.ast.Stmt([
            # __storetemp__(_it, __loadattr__(__builtins__, iter)(<list>))
            compiler.ast.CallFunc(special_name("__storetemp__"), [
                temp,
                compiler.ast.CallFunc(
                    compiler.ast.CallFunc(special_name("__loadattr__"),
                        [special_name("__builtins__"), special_name("iter")]
                        ),
                    [self.dispatch(node.list)]
                    )
                ]),
            # while True: ...
            compiler.ast.While(
                special_name("True"),
                # try: ...
                compiler.ast.TryExcept(
                    compiler.ast.Stmt([
                        # <assign> = ...
                        self.dispatch(node.assign,
                            # _it.next()
                            compiler.ast.CallFunc(
                                compiler.ast.CallFunc(special_name("__loadattr__"), [
                                    compiler.ast.CallFunc(special_name("__loadtemp__"), [temp]),
                                    special_name("next")
                                    ]),
                                []
                                )
                            )
                        ]),
                    # except StopIteration: ...
                    [(special_name("StopIteration"), None, compiler.ast.Stmt([compiler.ast.Break()]))],
                    # else: ...
                    body
                    ),
                None
                )
            ])

    def visitListCompIf(self, node, out_temp, expr, quals):

        # Either generate more "for" or "if" constructs.

        if quals:
            body = self.dispatch(quals[0], out_temp, expr, quals[1:])

        # Or generate the append statement.

        else:
            body = self._visitListCompExpr(out_temp, expr)

        return compiler.ast.If(
            [(self.dispatch(node.test), body)],
            None
            )

    def _visitListCompExpr(self, out_temp, expr):

        "To the 'out_temp' object, append the result of 'expr'."

        return compiler.ast.Stmt([
            # _out.append(<expr>)
            compiler.ast.CallFunc(
                compiler.ast.CallFunc(special_name("__loadattr__"), [
                    compiler.ast.CallFunc(special_name("__loadtemp__"), [out_temp]),
                    special_name("append")
                    ]),
                [self.dispatch(expr)]
                )
            ])

    def visitMod(self, node):
        return self._visitBinary(node)

    def visitMul(self, node):
        return self._visitBinary(node)

    def visitName(self, node, expr=None):
        if expr:
            return self.visitAssName(node, expr)

        unit = self.get_unit()
        attr = node._attr
        scope = node._scope

        # Generate appropriate name access operation.

        if scope == "constant":
            return node

        # Function locals are referenced normally.

        elif scope == "local" and isinstance(unit, Function):
            return node

        # Other attributes should already be resolved.

        elif attr is not None and not isinstance(attr, Instance):
            return compiler.ast.CallFunc(
                special_name("__loadattr__"),
                [quoted_ref(attr.parent), special_name(node.name)]
                )

        # Function globals are referenced via the module.

        elif scope == "global":
            return compiler.ast.CallFunc(
                special_name("__loadattr__"),
                [quoted_ref(self.get_module()), special_name(node.name)]
                )

        # NOTE: Unknown attributes may arise because a class attribute has been
        # NOTE: optimised away.

        else:
            return compiler.ast.CallFunc(
                special_name("__loadunknown__"),
                [special_name(node.name)]
                )

    def visitNot(self, node):
        return compiler.ast.Not(self.dispatch(node.expr))

    def visitOr(self, node):
        return compiler.ast.Or([self.dispatch(n) for n in node.nodes])

    def visitPower(self, node):
        return self._visitBinary(node)

    def visitRightShift(self, node):
        return self._visitBinary(node)

    def visitSlice(self, node, expr=None):
        return compiler.ast.CallFunc(
            module_attribute("operator", expr and "setslice" or "getslice"),
            [self.dispatch(node.expr), node.lower and self.dispatch(node.lower), node.upper and self.dispatch(node.upper)]
            + (expr and [expr] or [])
            )

    def visitSliceobj(self, node):
        return compiler.ast.Sliceobj([self.dispatch(n) for n in node.nodes])

    def visitSub(self, node):
        return self._visitBinary(node)

    def visitSubscript(self, node, expr=None):
        return compiler.ast.CallFunc(
            module_attribute("operator", expr and "setitem" or "getitem"),
            [self.dispatch(node.expr), compiler.ast.Tuple([self.dispatch(sub) for sub in node.subs])]
            + (expr and [expr] or [])
            )

    def visitTuple(self, node, expr=None):
        if expr:
            return self.visitAssTuple(node, expr)
        return compiler.ast.Tuple([self.dispatch(n) for n in node.nodes])

    def visitUnaryAdd(self, node):
        return self._visitUnary(node)

    def visitUnarySub(self, node):
        return self._visitUnary(node)

# Convenience functions.

def convert(module, program, filename):
    stream = open(filename, "wb")
    try:
        source = ConvertedSource(module, program)
        source.to_stream(stream)
    finally:
        stream.close()

def translate(program, directory):
    if not exists(directory):
        os.mkdir(directory)

    # NOTE: Add constants here.

    for module in program.get_importer().get_modules():
        convert(module, program, join(directory, "%s%spy" % (module.full_name(), extsep)))

# vim: tabstop=4 expandtab shiftwidth=4
